<?php

namespace Drupal\db_text_searcher\Commands;

use Drupal\Core\Database\Connection;
use Drupal\Core\File\FileSystemInterface;
use Drush\Commands\DrushCommands;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides Drush commands for conducting text searches across database tables.
 */
class SearchTextCommands extends DrushCommands {

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected $fileSystem;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The request stack service.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * Constructor for WebformSubmissionCommands.
   *
   * Sets up the command with necessary services.
   *
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service, used for file operations.
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection, used for database queries.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   The request stack, which holds the master and sub-request objects,
   *   allowing access to request data anywhere in the program.
   */
  public function __construct(FileSystemInterface $file_system, Connection $database, RequestStack $requestStack) {
    parent::__construct();
    $this->fileSystem = $file_system;
    $this->database = $database;
    $this->requestStack = $requestStack;
  }

  /**
   * Searches for specific text across database tables.
   *
   * It will save any results into a file.
   *
   * @param string $searchText
   *   Text to search across the database.
   *
   * @command db-text-searcher:search
   * @aliases db-text-search
   * @usage db-text-search "sample text"
   *   Executes partial match search for "sample text".
   */
  public function search($searchText) {
    // Execute partial text match search across tables.
    $searchResults = $this->searchPartialMatch($searchText);

    if ($searchResults) {
      // Display search results in console and save to file.
      $this->printAndSaveResults($searchResults, $searchText);
    }
    else {
      // Display message if no results are found.
      $this->output()->writeln("No results found.");
    }
  }

  /**
   * Executes partial text search in eligible database columns.
   *
   * This function omits certain table types such as 'batch',
   * 'revision', or 'cache', and performs text matching for
   * the specified search text.
   *
   * @param string $searchText
   *   The text to search within the database.
   *
   * @return array
   *   An array of matching results with details including
   *   table name, column name, entity ID, and matched value.
   */
  private function searchPartialMatch($searchText) {
    $searchResults = [];
    $uniqueIdentifiersSeen = [];
    $batchSize = 1000;
    $totalBatchesProcessed = 0;

    try {
      // Encode the search text in JSON format to process JSON-like structures.
      $searchTextJson = json_encode($searchText);

      // Fetch all relevant tables, excluding certain types like.
      $schema_name = $this->database->getConnectionOptions()['database'];
      $query = $this->database->select('information_schema.tables', 't');
      $query->addField('t', 'table_name');
      $query->condition('t.table_schema', $schema_name, '=');
      $query->where("t.table_name NOT REGEXP 'batch|revision|cache|tmgmt'");
      $eligibleTables = $query->execute()->fetchAllKeyed(0, 0);

      // Iterate over each eligible table for text searching within columns.
      foreach ($eligibleTables as $table) {
        $columnDetails = $this->getTableColumnData($table);
        $primaryKeys = $columnDetails['primaryKeys'];
        $textColumns = $columnDetails['textColumns'];

        // Text search within each column.
        foreach ($textColumns as $column) {
          $columnCharacteristics = $this->getColumnCharacteristics($table, $column);
          $charset = $columnCharacteristics['charset'];

          // Adjust search text based on column's character set.
          $searchTextForQuery = ($charset === 'utf8' || $charset === 'utf8mb4')
            ? $searchText : iconv('UTF-8', 'ASCII//IGNORE', $searchText);
          $searchTextJsonForQuery = ($charset === 'utf8' || $charset === 'utf8mb4')
            ? str_replace(['\\/', '\\\\'], ['\\\\/', '\\\\\\'], $searchTextJson)
            : iconv('UTF-8', 'ASCII//IGNORE', $searchTextJson);
          $searchTextJsonForQuery = trim($searchTextJsonForQuery, '"');

          // Prepare query with conditions for text search.
          $resultQuery = $this->database->select($table, 't')
            ->fields('t', array_merge($primaryKeys, [$column]));
          $orGroup = $resultQuery->orConditionGroup()
            ->condition($column, '%' . $this->database->escapeLike($searchTextForQuery) . '%', 'LIKE')
            ->condition($column, '%' . $this->database->escapeLike($searchTextJsonForQuery) . '%', 'LIKE');
          $resultQuery->condition($orGroup);

          $offset = 0;
          $batchCount = 0;
          while (TRUE) {
            $limitedQuery = clone $resultQuery;
            $limitedQuery->range($offset, $batchSize);
            $result = $limitedQuery->execute();
            $resultRows = $result->fetchAll();

            if (empty($resultRows)) {
              break;
            }

            // Process and accumulate unique search results.
            foreach ($resultRows as $row) {
              $matchedContents = $this->extractMatchContext($row->$column, $searchTextForQuery);
              foreach ($matchedContents as $matchedContent) {
                $entityIdentifier = implode('-', array_map(fn($primaryKey) => $row->$primaryKey, $primaryKeys));
                $resultKey = "{$table}-{$entityIdentifier}-" . md5($matchedContent);

                if (!isset($uniqueIdentifiersSeen[$resultKey])) {
                  $uniqueIdentifiersSeen[$resultKey] = TRUE;
                  $searchResults[] = [
                    'table' => $table,
                    'column' => $column,
                    'entity' => $entityIdentifier,
                    'matched_value' => $matchedContent,
                  ];
                }
              }
            }

            $batchCount++;
            $totalBatchesProcessed++;
            $offset += $batchSize;
          }

          if ($this->output()->isVeryVerbose()) {
            $this->output()->writeln("Completed processing {$batchCount} batches for column '{$column}' in table '{$table}'.");
          }
        }
      }

      if ($this->output()->isVeryVerbose()) {
        $this->output()->writeln("Total batches processed across all tables: {$totalBatchesProcessed}.");
      }
      return $searchResults;
    }
    catch (\Exception $e) {
      $this->logger()->error('An error occurred during the search: @message', ['@message' => $e->getMessage()]);
      return [];
    }
  }

  /**
   * Links table name patterns to URL paths for different entities.
   *
   * This array maps partial names of database tables to the URL paths of the
   * entities stored in them. The keys are fragments of table names and the
   * values are the corresponding URL paths, which include a placeholder for
   * the entity ID.
   *
   * Example:
   * Tables containing 'node_' correspond to URLs starting with '/node/', so
   * entities in these tables can be accessed directly via their IDs in the URL.
   *
   * @var array<string, string> Maps table patterns to URL paths
   */
  protected $entityUrlPaths = [
    'node_' => '/node/{entity_id}',
    'taxonomy_term_' => '/taxonomy/term/{entity_id}',
    'user_' => '/user/{entity_id}',
    'menu_link_content' => '/admin/structure/menu/item/{entity_id}',
    'block_content_' => '/block/{entity_id}',
    'media_' => '/media/{entity_id}/edit',
    'redirect' => '/admin/config/search/redirect/edit/{entity_id}',
  ];

  /**
   * Outputs search results to the console and stores them in a CSV file.
   *
   * It prints details including the URLs for direct entity access in
   * the terminal, and saves the details into a timestamped CSV file.
   * This file will include table names, column names, entity IDs, URLs,
   * and snippets of the content that matched the search term.
   *
   * @param array $searchResults
   *   An array of search results. Each result contains details
   *   like table name, column name, entity ID, and matched text,
   *   along with a URL for direct access.
   * @param string $searchTerm
   *   The search term used to generate these results.
   */
  private function printAndSaveResults(array $searchResults, $searchTerm) {
    $this->output()->writeln("Search results for \"$searchTerm\":\n");

    $currentDateTime = date('Y-m-d_H-i-s');
    $uniqueHash = sha1(rand());
    $fileID = substr($uniqueHash, 0, 6);
    $csvFilePath = "public://text_search_results_{$currentDateTime}_{$fileID}.csv";
    $baseURL = $this->getBaseUrl();

    // Creating a temporary file for writing CSV data.
    $temporaryFilePath = 'temporary://temp_csv_file.csv';
    $fileHandle = fopen($temporaryFilePath, 'w+');
    if (!$fileHandle) {
      $this->logger()->error("Unable to open temporary file for writing.");
      return;
    }

    // Define column headers for the CSV file.
    fputcsv($fileHandle, ['Table', 'Column', 'Entity', 'URL', 'Matched Value']);

    // Process each result found.
    foreach ($searchResults as &$result) {
      $entityUrlPath = $this->resolveEntityUrl($result['table'], $result['entity']);
      $resolvedUrl = $entityUrlPath ? "{$baseURL}{$entityUrlPath}" : "[No direct URL]";

      $matchedValues = is_array($result['matched_value']) ? $result['matched_value'] : [$result['matched_value']];
      foreach ($matchedValues as $matchedValue) {
        $truncatedMatchedValue = strlen($matchedValue) > 50 ? substr($matchedValue, 0, 50) . '...' : $matchedValue;

        // Verbose logging, if enabled.
        if ($this->output()->isVerbose()) {
          $this->output()->writeln("Table: {$result['table']}, Column: {$result['column']},
                    Entity ID: {$result['entity']}, URL: {$resolvedUrl}, Matched: {$truncatedMatchedValue}");
        }

        // Recording data into the CSV file.
        fputcsv($fileHandle, [
          $result['table'],
          $result['column'],
          $result['entity'],
          $resolvedUrl,
          $matchedValue,
        ]);
      }
    }

    fclose($fileHandle);

    // Move the temporary CSV file to the final destination.
    try {
      $finalResultsUri = $this->fileSystem->move($temporaryFilePath, $csvFilePath, FileExists::Replace);
      if ($finalResultsUri) {
        // Fetch the full absolute server path from Drupal.
        $fullPath = $this->fileSystem->realpath($finalResultsUri);

        // Convert the full server path to a web-accessible URL.
        // Adjust according to your Drupal setup.
        $relativeWebPath = str_replace(DRUPAL_ROOT, '', $fullPath);

        // Construct the full downloadable URL.
        $downloadLink = $baseURL . $relativeWebPath;

        $this->output()->writeln("Results saved to: $downloadLink");
      }
      else {
        throw new \Exception("Failed to move the CSV file to the final destination.");
      }
    }
    catch (\Exception $e) {
      $this->logger()->error("Failed to write results file: " . $e->getMessage());
    }
  }

  /**
   * Gets the base URL of the current website.
   *
   * @return string
   *   The website's base URL for the current request.
   */
  protected function getBaseUrl() {
    $request = $this->requestStack->getCurrentRequest();
    return $request->getSchemeAndHttpHost();
  }

  /**
   * Generates a URL for an entity based on its table name and identifier.
   *
   * This function attempts to find a URL pattern that matches the given table
   * name. If a pattern is found, it replaces a placeholder within the pattern
   * with the entity's identifier or its parent's identifier, and returns the
   * complete URL. If no pattern is found for the table, it returns null.
   *
   * @param string $tableName
   *   The table name within the database.
   * @param int|string $entityId
   *   The identifier of the entity for which to generate the URL.
   *
   * @return string|null
   *   The resolved URL if available; otherwise, null.
   */
  private function resolveEntityUrl($tableName, $entityId) {
    // Specific handle for entities within the 'cohesion_layout_field_data'.
    if ($tableName === 'cohesion_layout_field_data') {
      // Selects the 'parent_id' from the 'cohesion_layout_field_data' table
      // where the identifier matches.
      $query = $this->database->select($tableName, 'table')
        ->fields('table', ['parent_id'])
        ->condition('id', $entityId)
        ->range(0, 1);

      // Execute query and fetch results.
      $result = $query->execute();
      $record = $result->fetch();

      if ($record) {
        // If a record is found, use 'parent_id' instead of the 'entityId'.
        $entityId = $record->parent_id;
      }
      else {
        // Return null if no record exists matching the specified entityId.
        return NULL;
      }
    }

    // Search for a URL pattern that matches the table name.
    foreach ($this->entityUrlPaths as $patternKey => $urlPattern) {
      if (str_contains($tableName, $patternKey)) {
        // If a matching key is found in the table name, replace the placeholder
        // in the URL pattern.
        return str_replace('{entity_id}', $entityId, $urlPattern);
      }
    }

    // Return null if no relevant URL pattern is found for this table.
    return NULL;
  }

  /**
   * Extracts context around a search term from text.
   *
   * Attempts decoding if content is JSON.
   * Checks both normal and JSON-encoded text.
   *
   * @param string $fullText
   *   The full text retrieved from a database entry.
   * @param string $searchTerm
   *   The keyword or phrase used for finding contextual matches.
   *
   * @return array
   *   The extracted contextual snippets or URLs, or an explanatory message.
   */
  private function extractMatchContext($fullText, $searchTerm) {
    // Decode and clean the full text.
    $cleanedText = html_entity_decode($fullText);
    $cleanedText = preg_replace('/\\\\+/', '', $cleanedText);
    $cleanedText = preg_replace('/[\x00-\x09\x0B\x0C\x0E-\x1F\x7F]/', '', $cleanedText);

    // Convert text to lowercase for case-insensitive matching.
    $cleanedText = strtolower($cleanedText);
    $normalizedSearchTerm = strtolower($searchTerm);

    // Collection to store unique URLs.
    $uniqueUrls = [];

    // Modified regex pattern matching both absolute and relative URLs.
    $urlRegexPattern = '/\b(?:https?:\/\/[^\s<>"{}|\^`\[\]]+|(?:[\w.-]+\/)+[\w.-]+(?:\/[\w.-]*)*)/i';

    // Extract all URLs found in text.
    if (preg_match_all($urlRegexPattern, $cleanedText, $urlMatches)) {
      foreach ($urlMatches[0] as $url) {
        $sanitizedUrl = trim(urldecode($url));
        if (str_contains($sanitizedUrl, $normalizedSearchTerm)) {
          $uniqueUrls[$sanitizedUrl] = TRUE;
        }
      }
    }

    // Prepare the text for further non-URL related searches.
    $cleanedText = preg_replace($urlRegexPattern, '', $cleanedText);
    $cleanedTextLength = strlen($cleanedText);

    // Collection to store unique context snippets.
    $uniqueContexts = [];
    $contextSearchRadius = 100;
    $searchPosition = 0;

    // Extract the context around the search term within the text.
    while (($searchPosition = strpos($cleanedText, $normalizedSearchTerm, $searchPosition)) !== FALSE) {
      $contextStart = max(0, $searchPosition - $contextSearchRadius);
      $contextEnd = min($cleanedTextLength, $searchPosition + strlen($normalizedSearchTerm) + $contextSearchRadius);
      $contextLength = $contextEnd - $contextStart;

      // Determine if ellipses should be prepended or appended
      // based on context spill over.
      $prefix = $contextStart > 0 ? '...' : '';
      $suffix = $contextEnd < $cleanedTextLength ? '...' : '';

      $contextSnippet = $prefix . substr($cleanedText, $contextStart, $contextLength) . $suffix;

      // Clean up the extracted context snippet.
      $contextSnippet = mb_convert_encoding($contextSnippet, 'UTF-8', 'UTF-8');
      $contextSnippet = preg_replace('/[\x00-\x09\x0B\x0C\x0E-\x1F\x7F]/', '', $contextSnippet);
      $contextSnippet = htmlspecialchars_decode($contextSnippet, ENT_QUOTES);

      $uniqueContexts[$contextSnippet] = TRUE;
      $searchPosition += strlen($normalizedSearchTerm);
    }

    // Merge results and return the unique collection of URLs
    // and context snippets.
    return array_keys(array_merge($uniqueContexts, $uniqueUrls));
  }

  /**
   * Retrieves primary key and text column metadata for a specified table.
   *
   * This function queries the information schema to determine which columns
   * in the specified table are primary keys and which are textual columns
   * (such as char, varchar, text, longtext).
   *
   * @param string $tableName
   *   The name of the table for which to retrieve column data.
   *
   * @return array
   *   An associative array containing two keys:
   *    - 'primaryKeys' => columns as primary keys,
   *    - 'textColumns' => textual data type columns.
   *
   *   Example return:
   *   [
   *     'primaryKeys' => ['id', 'user_id'],
   *     'textColumns' => ['username', 'email_address']
   *   ]
   */
  private function getTableColumnData($tableName) {
    // Check if the table exists in the current database.
    $schema_name = $this->database->getConnectionOptions()['database'];
    $query = $this->database->select('information_schema.tables', 't');
    $query->fields('t', ['table_name']);
    $query->condition('t.table_schema', $schema_name);
    $query->condition('t.table_name', $tableName);

    $validTables = $query->execute()->fetchAllKeyed(0, 0);

    // Verify if the provided table name exists in the valid tables list.
    if (!array_key_exists($tableName, $validTables)) {
      throw new \InvalidArgumentException("Invalid or unknown table: {$tableName}");
    }

    // Dynamic query for column details.
    $columnQuery = $this->database->select('information_schema.columns', 'c');
    $columnQuery->fields('c', ['column_name']);
    $columnQuery->condition('c.table_schema', $schema_name);
    $columnQuery->condition('c.table_name', $tableName);

    // Exclude columns that may match patterns for 'deleted' or 'delta'.
    $columnQuery->condition('c.column_name', '%' . $this->database->escapeLike('deleted') . '%', 'NOT LIKE');
    $columnQuery->condition('c.column_name', '%' . $this->database->escapeLike('delta') . '%', 'NOT LIKE');
    $columnQuery->condition('c.column_name', '%' . $this->database->escapeLike('langcode') . '%', 'NOT LIKE');

    // Add expressions for primary key and textual column detection.
    $columnQuery->addExpression("CASE WHEN column_key = 'PRI' THEN 1 ELSE 0 END", 'is_primary');
    $columnQuery->addExpression("CASE WHEN data_type IN ('char', 'varchar', 'text', 'longtext') THEN 1 ELSE 0 END", 'is_text');

    // Execute the query.
    $columnResult = $columnQuery->execute();

    // Initialize containers for primary key columns and textual columns.
    $primaryKeyColumns = [];
    $textualColumns = [];

    // Sort columns into primary key or text types based on their attributes.
    foreach ($columnResult as $columnData) {
      if ($columnData->is_primary) {
        $primaryKeyColumns[] = $columnData->column_name;
      }
      if ($columnData->is_text) {
        $textualColumns[] = $columnData->column_name;
      }
    }

    // Return an organized associative array of column categorization.
    return ['primaryKeys' => $primaryKeyColumns, 'textColumns' => $textualColumns];
  }

  /**
   * Retrieves character set and properties of a specified database column.
   *
   * @param string $tableName
   *   The database table name.
   * @param string $columnName
   *   The column name to inspect.
   *
   * @return array
   *   An array containing column characteristics, such as character set.
   */
  private function getColumnCharacteristics($tableName, $columnName) {
    $schema_name = $this->database->getConnectionOptions()['database'];

    $query = $this->database->select('information_schema.columns', 'c');
    $query->fields('c', ['character_set_name']);
    $query->condition('c.table_schema', $schema_name);
    $query->condition('c.table_name', $tableName);
    $query->condition('c.column_name', $columnName);

    // Run the query and grab the result.
    $result = $query->execute()->fetch();

    // Return the value of 'character_set_name' or default to utf8.
    return [
      'charset' => $result ? $result->character_set_name : 'utf8',
    ];
  }

}
